#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributers.

from fenrirscreenreader.core import debug
from fenrirscreenreader.core.i18n import _
from collections import Counter


class AttributeManager():
    def __init__(self):
        self.currAttributes = None
        self.prevAttributes = None
        self.currAttributeDelta = ''
        self.currAttributeCursor = None
        self.prefAttributeCursor = None
        self.init_default_attributes()
        self.prevLastCursorAttribute = None
        self.currLastCursorAttribute = None

    def initialize(self, environment):
        self.env = environment

    def shutdown(self):
        pass

    def set_last_cursor_attribute(self, lastCursorAttribute):
        self.prevLastCursorAttribute = self.currLastCursorAttribute
        self.currLastCursorAttribute = lastCursorAttribute

    def reset_last_cursor_attribute(self):
        self.prevLastCursorAttribute = None
        self.currLastCursorAttribute = None

    def is_last_cursor_attribute_change(self):
        if self.prevLastCursorAttribute is None:
            return False
        return self.prevLastCursorAttribute != self.currLastCursorAttribute

    def get_curr_attribute_cursor(self):
        return self.currAttributeCursor

    def is_attribute_cursor_active(self):
        return self.currAttributeCursor is not None

    def is_attribute_change(self):
        if not self.prevAttributes:
            return False
        return self.currAttributes != self.prevAttributes

    def reset_attribute_all(self):
        self.reset_attribute_delta()
        self.reset_attribute_cursor()

    def get_attribute_delta(self):
        return self.currAttributeDelta

    def reset_attribute_delta(self):
        self.currAttributeDelta = ''

    def set_attribute_delta(self, currAttributeDelta):
        self.currAttributeDelta = currAttributeDelta

    def reset_attribute_cursor(self):
        self.currAttributeCursor = None
        self.prefAttributeCursor = None

    def set_attribute_cursor(self, currAttributeCursor):
        self.prefAttributeCursor = self.currAttributeCursor
        self.currAttributeCursor = currAttributeCursor.copy()

    def reset_attributes(self, currAttributes):
        self.prevAttributes = None
        self.currAttributes = currAttributes

    def set_attributes(self, currAttributes):
        self.prevAttributes = self.currAttributes
        self.currAttributes = currAttributes.copy()

    def get_attribute_by_xy(self, x, y):
        if not self.currAttributes:
            return None
        if len(self.currAttributes) < y:
            return None
        if len(self.currAttributes[y]) < x - 1:
            return None
        try:
            return self.currAttributes[y][x]
        except KeyError:
            try:
                return self.defaultAttributes[0]
            except Exception as e:
                self.env['runtime']['DebugManager'].write_debug_out(
                    'AttributeManager get_attribute_by_xy: Error accessing default attributes: ' +
                    str(e),
                    debug.DebugLevel.ERROR)
        return None

    def append_default_attributes(self, attribute):
        if not attribute:
            return
        if len(attribute) != 10:
            return
        self.defaultAttributes.append(attribute)

    def init_default_attributes(self):
        self.defaultAttributes = [None]
        self.defaultAttributes.append([
            'default',  # fg
            'default',  # bg
            False,  # bold
            False,  # italics
            False,  # underscore
            False,  # strikethrough
            False,  # reverse
            False,  # blink
            'default',  # fontsize
            'default'  # fontfamily
        ])  # end attribute

    def is_default_attribute(self, attribute):
        return attribute in self.defaultAttributes

    def has_attributes(self, cursor, update=True):
        if not cursor:
            return False
        cursor_pos = cursor.copy()
        try:
            attribute = self.get_attribute_by_xy(cursor_pos['x'], cursor_pos['y'])

            if update:
                self.set_last_cursor_attribute(attribute)
            if not self.is_last_cursor_attribute_change():
                return False

            if self.is_default_attribute(attribute):
                return False

        except Exception as e:
            return False
        return True

    def format_attributes(self, attribute, attribute_format_string=''):
        # "black",
        # "red",
        # "green",
        # "brown",
        # "blue",
        # "magenta",
        # "cyan",
        # "white",
        # "default" # white.
        # _order_
        # "fg",
        # "bg",
        # "bold",
        # "italics",
        # "underscore",
        # "strikethrough",
        # "reverse",
        # "blink"
        # "fontsieze"
        # "fontfamily"
        if attribute_format_string == '':
            attribute_format_string = self.env['runtime']['SettingsManager'].get_setting(
                'general', 'attribute_format_string')
        if not attribute_format_string:
            return ''
        if attribute_format_string == '':
            return ''
        if not attribute:
            return ''
        if len(attribute) != 10:
            return ''

        # 0 FG color (name)
        try:
            attribute_format_string = attribute_format_string.replace(
                'fenrirFGColor', _(attribute[0]))
        except Exception as e:
            attribute_format_string = attribute_format_string.replace(
                'fenrirFGColor', '')

        # 1 BG color (name)
        try:
            attribute_format_string = attribute_format_string.replace(
                'fenrirBGColor', _(attribute[1]))
        except Exception as e:
            attribute_format_string = attribute_format_string.replace(
                'fenrirBGColor', '')

        # 2 bold (True/ False)
        try:
            if attribute[2]:
                attribute_format_string = attribute_format_string.replace(
                    'fenrirBold', _('bold'))
        except Exception as e:
            pass
        attribute_format_string = attribute_format_string.replace('fenrirBold', '')

        # 3 italics (True/ False)
        try:
            if attribute[3]:
                attribute_format_string = attribute_format_string.replace(
                    'fenrirItalics', _('italic'))
        except Exception as e:
            pass
        attribute_format_string = attribute_format_string.replace(
            'fenrirItalics', '')

        # 4 underline (True/ False)
        try:
            if attribute[4]:
                attribute_format_string = attribute_format_string.replace(
                    'fenrirUnderline', _('underline'))
        except Exception as e:
            pass
        attribute_format_string = attribute_format_string.replace(
            'fenrirUnderline', '')

        # 5 strikethrough (True/ False)
        try:
            if attribute[5]:
                attribute_format_string = attribute_format_string.replace(
                    'fenrirStrikethrough', _('strikethrough'))
        except Exception as e:
            pass
        attribute_format_string = attribute_format_string.replace(
            'fenrirStrikethrough', '')

        # 6 reverse (True/ False)
        try:
            if attribute[6]:
                attribute_format_string = attribute_format_string.replace(
                    'fenrirReverse', _('reverse'))
        except Exception as e:
            pass
        attribute_format_string = attribute_format_string.replace(
            'fenrirReverse', '')

        # 7 blink (True/ False)
        try:
            if attribute[7]:
                attribute_format_string = attribute_format_string.replace(
                    'fenrirBlink', _('blink'))
        except Exception as e:
            pass
        attribute_format_string = attribute_format_string.replace(
            'fenrirBlink', '')

        # 8 font size (int/ string)
        try:
            try:
                attribute_format_string = attribute_format_string.replace(
                    'fenrirFontSize', int(attribute[8]))
            except Exception as e:
                self.env['runtime']['DebugManager'].write_debug_out(
                    'AttributeManager formatAttributeToString: Error formatting font size as int: ' +
                    str(e),
                    debug.DebugLevel.ERROR)
            try:
                attribute_format_string = attribute_format_string.replace(
                    'fenrirFontSize', str(attribute[8]))
            except Exception as e:
                self.env['runtime']['DebugManager'].write_debug_out(
                    'AttributeManager formatAttributeToString: Error formatting font size as string: ' +
                    str(e),
                    debug.DebugLevel.ERROR)
        except Exception as e:
            pass
        attribute_format_string = attribute_format_string.replace(
            'fenrirFontSize', _('default'))

        # 9 font family (string)
        try:
            attribute_format_string = attribute_format_string.replace(
                'fenrirFont', attribute[9])
        except Exception as e:
            pass
        attribute_format_string = attribute_format_string.replace(
            'fenrirFont', _('default'))

        return attribute_format_string

    def track_highlights(self):
        """
        Detects text with changed attributes (highlighting) between screen updates.

        This is crucial for screen readers to announce when text becomes highlighted,
        selected, or changes visual emphasis (bold, reverse video, color changes, etc.)

        Returns:
            tuple: (highlighted_text, cursor_position)
                - highlighted_text: string of characters that gained highlighting
                - cursor_position: dict {'x': col, 'y': row} of first highlighted char
        """
        result = ''
        curr_cursor = None

        # Early exit conditions - no attribute comparison possible
        if self.prevAttributes is None:
            # First screen load - no previous attributes to compare against
            return result, curr_cursor
        if self.prevAttributes == self.currAttributes:
            # No attribute changes detected
            return result, curr_cursor
        if self.currAttributes is None:
            # Error condition - current attributes missing
            return result, curr_cursor
        if len(self.currAttributes) == 0:
            # Special case for PTY environments with no text content
            return result, curr_cursor

        # Get current screen text to correlate with attribute changes
        text = self.env['runtime']['ScreenManager'].get_screen_text()
        text_lines = text.split('\n')

        # Sanity check: text lines must match attribute array dimensions
        if len(text_lines) != len(self.currAttributes):
            return result, curr_cursor

        # Compare attributes line by line, character by character
        for line in range(len(self.prevAttributes)):
            if self.prevAttributes[line] != self.currAttributes[line]:
                # This line has attribute changes - examine each character
                # position
                for column in range(len(self.prevAttributes[line])):
                    if self.prevAttributes[line][column] == self.currAttributes[line][column]:
                        # No change at this position
                        continue
                    # Attribute changed at this position - check if it's worth
                    # announcing
                    if self.is_useful_for_tracking(
                            line, column, self.currAttributes, self.prevAttributes):
                        # First highlighted character becomes cursor position
                        # for navigation
                        if not curr_cursor:
                            curr_cursor = {'x': column, 'y': line}
                        # Accumulate highlighted characters
                        result += text_lines[line][column]
                # Add space between lines of highlighted text for speech
                # clarity
                result += ' '
        return result, curr_cursor

    def is_useful_for_tracking(
            self,
            line,
            column,
            currAttributes,
            prevAttributes,
            attribute=1,
            mode='zaxe'):
        """
        Determines if an attribute change at a specific position is worth announcing.

        This prevents announcing every minor attribute change and focuses on meaningful
        highlighting that indicates user selections, focus changes, or important emphasis.

        Args:
            line, column: Position of the attribute change
            currAttributes, prevAttributes: Current and previous attribute arrays
            attribute: Which attribute to examine (1=background color by default)
            mode: Detection algorithm ('zaxe', 'default', 'barrier')

        Returns:
            bool: True if this attribute change should be announced to user
        """
        # Sanity checks for valid position and sufficient screen content
        if len(currAttributes) <= 3:
            return False
        if line < 0:
            return False
        if line > len(currAttributes):
            return False

        useful = False

        if mode == 'default':
            # Simple mode: announce any non-default attribute
            useful = not self.is_default_attribute(currAttributes[line][column])

        elif (mode == 'zaxe') or (mode == ''):
            # Context-aware mode: only announce attributes that stand out from surroundings
            # This prevents announcing entire blocks of highlighted text character by character
            # by checking if the attribute differs from adjacent lines

            if line == 0:
                # Top line: compare against lines below
                useful = (currAttributes[line][column][attribute] != currAttributes[line + 1][column][attribute]) and (
                    currAttributes[line][column][attribute] != currAttributes[line + 2][column][attribute])
            elif line >= len(prevAttributes):
                # Bottom line: compare against lines above
                useful = (currAttributes[line][column][attribute] != currAttributes[line - 1][column][attribute]) and (
                    currAttributes[line][column][attribute] != currAttributes[line - 2][column][attribute])
            else:
                # Middle lines: compare against both directions
                useful = (currAttributes[line][column][attribute] != currAttributes[line + 1][column][attribute]) and (
                    currAttributes[line][column][attribute] != currAttributes[line - 1][column][attribute])

        elif mode == 'barrier':
            # Barrier mode: future enhancement for detecting screen
            # boundaries/separators
            useful = True

        return useful
