#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Fenrir TTY screen reader
# By Chrys, Storm Dragon, and contributors.

import signal
import time
import os
import sys

from fenrirscreenreader.core.i18n import _
from fenrirscreenreader.core import settingsManager
from fenrirscreenreader.core import debug
from fenrirscreenreader.core.eventData import FenrirEventType


class FenrirManager():
    def __init__(self, cliArgs):
        self.is_initialized = False
        try:
            self.environment = settingsManager.SettingsManager().init_fenrir_config(cliArgs, self)
            if not self.environment:
                raise RuntimeError(
                    'Cannot Initialize. Maybe the configfile is not available or not parseable')
        except RuntimeError:
            raise

        self.environment['runtime']['OutputManager'].present_text(
            _("Start Fenrir"), sound_icon ='ScreenReaderOn', interrupt=True)
        signal.signal(signal.SIGINT, self.capture_signal)
        signal.signal(signal.SIGTERM, self.capture_signal)

        self.is_initialized = True
        self.modifierInput = False
        self.singleKeyCommand = False
        self.command = ''
        self.set_process_name()

    def proceed(self):
        if not self.is_initialized:
            return
        self.environment['runtime']['EventManager'].start_main_event_loop()
        self.shutdown()

    def handle_input(self, event):
        self.environment['runtime']['DebugManager'].write_debug_out(
            'DEBUG INPUT fenrirMan:' + str(event), debug.DebugLevel.INFO)

        if not event['data']:
            event['data'] = self.environment['runtime']['InputManager'].get_input_event()

        if event['data']:
            event['data']['EventName'] = self.environment['runtime']['InputManager'].convert_event_name(
                event['data']['EventName'])
            self.environment['runtime']['InputManager'].handle_input_event(
                event['data'])
        else:
            return

        if self.environment['runtime']['InputManager'].no_key_pressed():
            self.environment['runtime']['InputManager'].clear_last_deep_input()

        if self.environment['runtime']['ScreenManager'].is_ignored_screen():
            self.environment['runtime']['InputManager'].write_event_buffer()
        else:
            if self.environment['runtime']['HelpManager'].is_tutorial_mode():
                self.environment['runtime']['InputManager'].clear_event_buffer()
                self.environment['runtime']['InputManager'].key_echo(
                    event['data'])

            if self.environment['runtime']['VmenuManager'].get_active():
                self.environment['runtime']['InputManager'].clear_event_buffer()

            self.detect_shortcut_command()

            if self.modifierInput:
                self.environment['runtime']['InputManager'].clear_event_buffer()
            if self.singleKeyCommand:
                if self.environment['runtime']['InputManager'].no_key_pressed():
                    self.environment['runtime']['InputManager'].clear_event_buffer()
            else:
                self.environment['runtime']['InputManager'].write_event_buffer()

        if self.environment['runtime']['InputManager'].no_key_pressed():
            self.modifierInput = False
            self.singleKeyCommand = False
            self.environment['runtime']['InputManager'].write_event_buffer()
            self.environment['runtime']['InputManager'].handle_device_grab()

        if self.environment['input']['keyForeward'] > 0:
            self.environment['input']['keyForeward'] -= 1

        self.environment['runtime']['CommandManager'].execute_default_trigger(
            'onKeyInput')

    def handle_byte_input(self, event):
        if not event['data'] or event['data'] == b'':
            return
        self.environment['runtime']['ByteManager'].handle_byte_input(
            event['data'])
        self.environment['runtime']['CommandManager'].execute_default_trigger(
            'onByteInput')

    def handle_execute_command(self, event):
        if not event['data'] or event['data'] == '':
            return
        current_command = event['data']

        # special modes
        if self.environment['runtime']['HelpManager'].is_tutorial_mode():
            if self.environment['runtime']['CommandManager'].command_exists(
                    current_command, 'help'):
                self.environment['runtime']['CommandManager'].execute_command(
                    current_command, 'help')
                return
        elif self.environment['runtime']['VmenuManager'].get_active():
            if self.environment['runtime']['CommandManager'].command_exists(
                    current_command, 'vmenu-navigation'):
                self.environment['runtime']['CommandManager'].execute_command(
                    current_command, 'vmenu-navigation')
                return

        # default
        self.environment['runtime']['CommandManager'].execute_command(
            current_command, 'commands')

    def handle_remote_incomming(self, event):
        if not event['data']:
            return
        self.environment['runtime']['remoteManager'].handle_remote_incomming(
            event['data'])

    def handle_screen_change(self, event):
        self.environment['runtime']['ScreenManager'].handle_screen_change(
            event['data'])
        if self.environment['runtime']['VmenuManager'].get_active():
            return
        self.environment['runtime']['CommandManager'].execute_default_trigger(
            'onScreenChanged')
        self.environment['runtime']['ScreenDriver'].get_curr_screen()

    def handle_screen_update(self, event):
        self.environment['runtime']['ScreenManager'].handle_screen_update(
            event['data'])

        if time.time() - \
                self.environment['runtime']['InputManager'].get_last_input_time() >= 0.3:
            self.environment['runtime']['InputManager'].clear_last_deep_input()

        if (self.environment['runtime']['CursorManager'].is_cursor_vertical_move(
        ) or self.environment['runtime']['CursorManager'].is_cursor_horizontal_move()):
            self.environment['runtime']['CommandManager'].execute_default_trigger(
                'onCursorChange')

        self.environment['runtime']['CommandManager'].execute_default_trigger(
            'onScreenUpdate')
        self.environment['runtime']['InputManager'].clear_last_deep_input()

    def handle_plug_input_device(self, event):
        try:
            self.environment['runtime']['InputManager'].set_last_detected_devices(
                event['data'])
        except Exception as e:
            self.environment['runtime']['DebugManager'].write_debug_out(
                'handle_plug_input_device: Error setting last detected devices: ' + str(e),
                debug.DebugLevel.ERROR)
        self.environment['runtime']['InputManager'].handle_plug_input_device(
            event['data'])
        self.environment['runtime']['CommandManager'].execute_default_trigger(
            'onPlugInputDevice', force=True)
        self.environment['runtime']['InputManager'].set_last_detected_devices(
            None)

    def handle_heart_beat(self, event):
        self.environment['runtime']['CommandManager'].execute_default_trigger(
            'onHeartBeat', force=True)

    def detect_shortcut_command(self):
        if self.environment['input']['keyForeward'] > 0:
            return

        if len(self.environment['input']['prevInput']) > len(
                self.environment['input']['currInput']):
            return

        if self.environment['runtime']['InputManager'].is_key_press():
            self.modifierInput = self.environment['runtime']['InputManager'].curr_key_is_modifier(
            )
        else:
            if not self.environment['runtime']['InputManager'].no_key_pressed():
                if self.singleKeyCommand:
                    self.singleKeyCommand = len(
                        self.environment['input']['currInput']) == 1

        if not (
                self.singleKeyCommand and self.environment['runtime']['InputManager'].no_key_pressed()):
            current_shortcut = self.environment['runtime']['InputManager'].get_curr_shortcut(
            )
            self.command = self.environment['runtime']['InputManager'].get_command_for_shortcut(
                current_shortcut)

        if not self.modifierInput:
            if self.environment['runtime']['InputManager'].is_key_press():
                if self.command != '':
                    self.singleKeyCommand = True

        if not (self.singleKeyCommand or self.modifierInput):
            return

        # fire event
        if self.command != '':
            if self.modifierInput:
                self.environment['runtime']['EventManager'].put_to_event_queue(
                    FenrirEventType.execute_command, self.command)
                self.command = ''
            else:
                if self.singleKeyCommand:
                    self.environment['runtime']['EventManager'].put_to_event_queue(
                        FenrirEventType.execute_command, self.command)
                    self.command = ''

    def set_process_name(self, name='fenrir'):
        """Attempts to set the process name to 'fenrir'."""
        try:
            from setproctitle import setproctitle
        except ImportError:
            pass
        else:
            setproctitle(name)
            return True

        try:
            from ctypes import cdll, byref, create_string_buffer
            libc = cdll.LoadLibrary('libc.so.6')
            string_buffer = create_string_buffer(len(name) + 1)
            string_buffer.value = bytes(name, 'UTF-8')
            libc.prctl(15, byref(string_buffer), 0, 0, 0)
            return True
        except Exception as e:
            self.environment['runtime']['DebugManager'].write_debug_out(
                'setProcName: Error setting process name: ' + str(e), debug.DebugLevel.ERROR)

        return False

    def shutdown_request(self):
        try:
            self.environment['runtime']['EventManager'].stop_main_event_loop()
        except Exception as e:
            self.environment['runtime']['DebugManager'].write_debug_out(
                'shutdown_request: Error stopping main event loop: ' + str(e),
                debug.DebugLevel.ERROR)

    def capture_signal(self, sigInit, frame):
        self.shutdown_request()

    def shutdown(self):
        self.environment['runtime']['InputManager'].ungrab_all_devices()
        self.environment['runtime']['EventManager'].stop_main_event_loop()
        self.environment['runtime']['OutputManager'].present_text(
            _("Quit Fenrir"), sound_icon ='ScreenReaderOff', interrupt=True)
        self.environment['runtime']['EventManager'].clean_event_queue()
        time.sleep(0.6)

        for currentManager in self.environment['general']['managerList']:
            if self.environment['runtime'][currentManager]:
                self.environment['runtime'][currentManager].shutdown()
                del self.environment['runtime'][currentManager]

        self.environment = None
